/*
 * grid.cpp: definition of the Grid class
 */

#include <stdlib.h>
#include <iostream.h>
#include <strstrea.h>
#include <fstream.h>

#include "element.h"
#include "index.h"
#include "randnum.h"
#include "grid.h"

/*
 * generate a 2D array of elements
 */
bool Grid::generateArray()
{
  // grab memory
  elem = new Element[idx.size()];
  if (!elem) {
    return false;
  }

  return true;
}

/*
 * add the addresses of neighbouring elements to each element
 * in the array - use 0 for address if at boundary
 */
void Grid::addNeighbours()
{
  // the idx object is of class Index and checks if we are 
  // asking for a position within the grid boundaries
  Index idx(ncols, nrows);

  Element *leftPtr, *rightPtr, *abovePtr, *belowPtr;

  for (int i=0; i<ncols; i++) {
    for (int j=0; j<nrows; j++) {
      leftPtr =  idx.isOK(i-1, j) ? &elem[idx(i-1, j)] : 0;
      rightPtr = idx.isOK(i+1, j) ? &elem[idx(i+1, j)] : 0;
      abovePtr = idx.isOK(i, j-1) ? &elem[idx(i, j-1)] : 0;
      belowPtr = idx.isOK(i, j+1) ? &elem[idx(i, j+1)] : 0;
      
      elem[idx(i, j)].addNeighbours(leftPtr, rightPtr, abovePtr, belowPtr);
    }
  }
}

/*
 * Construction & Destruction
 */
Grid::Grid(int _ncols, int _nrows)
: ncols(_ncols), nrows(_nrows), idx(_ncols, _nrows), elem(0)
{
  // generate the element array
  if (!generateArray()) {
    cerr << "Out of memory!\n";
    exit(-2);
  }

  // add the addresses of each element's neighbours
  addNeighbours();
}

Grid::~Grid()
{
  if (elem) {
    delete[] elem;
  }
}

/*
 * initialise the (u,v) values in each element
 * do it randomly here, but could be read in from file
 */
void Grid::init()
{
  // r is an instance of my class RandNum that seeds the random
  // number generator with the current time and then returns
  // a series of (pseudo) random numbers. Number return is between 0 and 1
  RandNum r;

  for (int i=0; i<ncols; i++) {
    for (int j=0; j<nrows; j++) {
      elem[idx(i, j)].init(r.next(), r.next());
    }
  }
}

/*
 * do a timestep by calling the 'update' method on each element
 * and them normalising the vector represented by (u,v)
 */
void Grid::step()
{
  int i, j;
  
  // do the calculation
  for (i=0; i<ncols; i++) {
    for (j=0; j<nrows; j++) {
      elem[idx(i, j)].update();
    }
  }
  // normalise vector (u,v) and compute 'result'
  for (i=0; i<ncols; i++) {
    for (j=0; j<nrows; j++) {
      Element* e = &elem[idx(i, j)];
      e->normalise();
      e->sumOfParts();
    }
  }
}

/*
 * output the magnitudes
 */
void Grid::output(int fileNumber)
{
  char outName[128];
  strstream str(outName, 128, ios::out);

  str << "grid" << fileNumber << ".dat" << '\0';

  ofstream out(outName);

  for (int i=0; i<ncols; i++) {
    for (int j=0; j<nrows; j++) {
      out << elem[idx(i, j)].result << " ";
    }
    out << "\n";
  }

  out.close();
}

/*
 * E N D  O F  F I L E
 */
